// Copyright (C) 2018 LAAS-CNRS
// Author: Pierre Fernbach
//
// This file is part of the hpp-rbprm.
//
// hpp-rbprm is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// test-hpp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with hpp-core.  If not, see <http://www.gnu.org/licenses/>.


#define BOOST_TEST_MODULE test-reachability
#include <boost/test/included/unit_test.hpp>

#include <hpp/rbprm/contact_generation/reachability.hh>
#include "tools-fullbody.hh"


using namespace hpp;
using namespace rbprm;




BOOST_AUTO_TEST_SUITE( reachability_quasiStatic )

BOOST_AUTO_TEST_CASE (kin_constraints_loaded) {

    RbPrmFullBodyPtr_t fullBody = loadHRP2();

    for(rbprm::CIT_Limb lit = fullBody->GetLimbs().begin() ; lit != fullBody->GetLimbs().end() ; ++lit){
        BOOST_CHECK(lit->second->kinematicConstraints_.first.rows() > 0);
        BOOST_CHECK(lit->second->kinematicConstraints_.second.rows() > 0);
        BOOST_CHECK_EQUAL(lit->second->kinematicConstraints_.first.rows() ,lit->second->kinematicConstraints_.second.rows());
    }
}

BOOST_AUTO_TEST_CASE(reachable_quasiStatic_rightFoot_front){
    RbPrmFullBodyPtr_t fullBody = loadHRP2();

    std::vector<std::string> rleg,lleg;
    rleg.push_back("hrp2_rleg_rom");
    lleg.push_back("hrp2_lleg_rom");


    core::Configuration_t q0(fullBody->device_->configSize()),q02(fullBody->device_->configSize()),q04(fullBody->device_->configSize()),q06(fullBody->device_->configSize()),q07(fullBody->device_->configSize()),q075(fullBody->device_->configSize()),q08(fullBody->device_->configSize());

    //configurations found in python, for a step with the right foot in a straight line of various length

    q0<<4.9999999999999996e-06,
            0.0,
            0.658702,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.0,
            0.0,
            -0.453785606,
            0.872664626,
            -0.41887902,
            0.0,
            0.0,
            0.0,
            -0.453785606,
            0.872664626,
            -0.41887902,
            0.0,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;
    q02<<0.08843182423782969,
            0.0011275296729917929,
            0.6548324227315108,
            -0.004885261937904553,
            -0.00489605086429045,
            0.01274546631763341,
            0.9998948524670372,
            -9.134404225263402e-17,
            1.042152998990706e-17,
            -1.9459076020781713e-17,
            8.868954565057905e-19,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.0255239017883892,
            0.007585126274920495,
            -0.28028005379552895,
            0.865578150843098,
            -0.5753811165825214,
            0.002059066459726205,
            -0.02542821341602173,
            -0.0019321466049652988,
            -0.6372450527738972,
            0.8855554244750417,
            -0.23839432492611554,
            0.011577280206887898,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q04<<0.18835767718866914,
            -0.003995082424102922,
            0.6281901711546837,
            0.002085909708105276,
            -0.0047520274380174355,
            0.02679189592611454,
            0.9996275614090491,
            -1.978205330344533e-16,
            2.1382084021341444e-17,
            2.134616917095093e-17,
            9.058543913216502e-19,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.05364475449588497,
            0.0035224416249896155,
            -0.10089804528299286,
            0.8794784596968881,
            -0.7691914710220066,
            -0.007947707016615997,
            -0.05354980395574721,
            -0.006594905751357978,
            -0.8169105276356746,
            0.866982298982374,
            -0.04068268406774709,
            0.0021700883852268384,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q06<<0.2905954544897296,
            -0.007170181849911604,
            0.5805865931676134,
            0.007401935462840584,
            -0.002181395566012728,
            0.04230833220613408,
            0.9990748009487198,
            -3.267618377782896e-16,
            2.5565025868505295e-17,
            -1.3981398426668165e-17,
            -6.6842514330955285e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.08467524128525662,
            0.0008598651163053476,
            0.0962675319004384,
            0.8824305824721005,
            -0.974965663884218,
            -0.015835298531193934,
            -0.08463299611046105,
            -0.010687153054445111,
            -1.009575442860765,
            0.8466952649613847,
            0.16661279638717294,
            -0.004288129241244868,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q07<<0.34189722745035456,
            -0.008012429178832478,
            0.5472670660731433,
            0.009581959244352473,
            -7.564560826158652e-05,
            0.05114924155851632,
            0.9986450497663171,
            -2.7399870945889274e-16,
            4.216193532779036e-17,
            -6.58893674960029e-18,
            -4.8696805704366236e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.10233874090853924,
            7.266136650437027e-05,
            0.20458808773339823,
            0.8780716412003806,
            -1.083488880863484,
            -0.019219405034467056,
            -0.10235178592233966,
            -0.012757906288768958,
            -1.1139320325950293,
            0.8333155010879875,
            0.279787165920026,
            -0.006388556670063253,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q075<<0.3675759436734773,
            -0.008230122400470406,
            0.5277578352254366,
            0.010560616076045517,
            0.0012976088082582754,
            0.056006415403234644,
            0.9983737131120567,
            -3.1652970926267093e-16,
            3.74116327140296e-17,
            4.3416408221931256e-18,
            -2.208004704127694e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.11203701719669751,
            -0.0001444598896007637,
            0.2621286574257498,
            0.8736792885091234,
            -1.1395819539777483,
            -0.02079840203942289,
            -0.11209257395795832,
            -0.013856625614772503,
            -1.168807733579881,
            0.824966455665839,
            0.34006665054904944,
            -0.007085623020689349,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;
    q08<<0.3932766434540968,
            -0.008299049022497723,
            0.506081064665376,
            0.011457923700630398,
            0.0029669929320899863,
            0.06125803516890408,
            0.9980517852619958,
            -3.792752430536026e-16,
            3.974406823517091e-17,
            -4.48131472602152e-20,
            1.4836031061741746e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            -0.12251747184831462,
            -0.00020203657791714367,
            0.3224647771125514,
            0.867299136509055,
            -1.1970903909199129,
            -0.0223073182050388,
            -0.12263197243441731,
            -0.015029070781306936,
            -1.2258425807324829,
            0.8150032254395925,
            0.40351160481696147,
            -0.007478952436920722,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;


    State s0 = createState(fullBody,q0);
    State s02 = createState(fullBody,q02);
    State s04 = createState(fullBody,q04);
    State s06 = createState(fullBody,q06);
    State s07 = createState(fullBody,q07);
    State s075 = createState(fullBody,q075);
    State s08 = createState(fullBody,q08);

    BOOST_CHECK(reachability::isReachable(fullBody,s0,s02).success());
    BOOST_CHECK(reachability::isReachable(fullBody,s0,s04).success());
    BOOST_CHECK(reachability::isReachable(fullBody,s0,s06).success());
    BOOST_CHECK(reachability::isReachable(fullBody,s0,s07).success());
    BOOST_CHECK( ! reachability::isReachable(fullBody,s0,s075).success());
    BOOST_CHECK( ! reachability::isReachable(fullBody,s0,s08).success()); // should fail in quasi-static

}


BOOST_AUTO_TEST_CASE(reachable_quasiStatic_rightFoot_backward){
    RbPrmFullBodyPtr_t fullBody = loadHRP2();


    std::vector<std::string> rleg,lleg;
    rleg.push_back("hrp2_rleg_rom");
    lleg.push_back("hrp2_lleg_rom");


    core::Configuration_t q0(fullBody->device_->configSize()),q02(fullBody->device_->configSize()),q04(fullBody->device_->configSize()),q06(fullBody->device_->configSize()),q07(fullBody->device_->configSize()),q075(fullBody->device_->configSize());

    //configurations found in python, for a step with the right foot in a straight line of various length

    q0<<4.9999999999999996e-06,
            0.0,
            0.658702,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.0,
            0.0,
            -0.453785606,
            0.872664626,
            -0.41887902,
            0.0,
            0.0,
            0.0,
            -0.453785606,
            0.872664626,
            -0.41887902,
            0.0,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q02<<-0.09758900019415818,
            0.005901017324075486,
            0.6541835446757959,
            -0.011415109957213333,
            0.0053054514727031935,
            -0.014157622738680858,
            0.9998205384806439,
            1.0041515222558578e-16,
            -1.489154233972501e-17,
            -1.6673878059977163e-17,
            1.9637172900298163e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.028327578538278962,
            0.011758918402547284,
            -0.6253279828851873,
            0.8415339824387751,
            -0.22713984003887597,
            0.010919071400508556,
            0.02821985193774157,
            0.0020876064554544824,
            -0.2714246370018223,
            0.8955981967623615,
            -0.6351058221318389,
            0.020591512778174517,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q04<<-0.1975688810371278,
            0.005393293849887926,
            0.6257930375123042,
            -0.011461280693882376,
            0.006032664803307026,
            -0.02818231339200397,
            0.9995188858706483,
            1.8442886254987655e-16,
            -2.473664487562888e-17,
            -8.564331426998748e-17,
            3.858326245241586e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.05638840969836136,
            0.012201394010970064,
            -0.8114909343108919,
            0.836327142794806,
            -0.03754303349182045,
            0.01037286209706479,
            0.05625473738612886,
            0.0016762614379669772,
            -0.08314608567512384,
            0.8945042546081007,
            -0.8240640633698866,
            0.02089884828703267,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q06<<-0.29903596704034463,
            0.0055265357029923795,
            0.5757702795178605,
            -0.013001010560369432,
            0.006667591359708588,
            -0.04378907270098682,
            0.9989339487983462,
            2.8134107462431333e-16,
            -2.5896162952480462e-17,
            -2.1229452527397024e-17,
            4.9481559633730675e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.0876323884714764,
            0.013899427738306564,
            -1.009749053812744,
            0.8223681459284908,
            0.1729194549128982,
            0.011494919409771938,
            0.08745571681165812,
            0.0016000852195584604,
            0.11735009322256851,
            0.8924275567476551,
            -1.024237696695266,
            0.02379562391383505,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q07<<-0.34965903133902637,
            0.005535314202306755,
            0.540921449010816,
            -0.01411407289508837,
            0.00781857593047554,
            -0.0527620111309932,
            0.9984767563634841,
            3.6812433790753916e-16,
            -4.365697802765814e-17,
            -5.932848182288139e-17,
            9.479967443503129e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.10561037820039874,
            0.015154853581334376,
            -1.1164774222632243,
            0.8101277583215767,
            0.28924426892516886,
            0.0122108012433152,
            0.10537733362790859,
            0.0013256584819744646,
            0.22704580766056906,
            0.8872962343186303,
            -1.1314454247052832,
            0.02604224579140171,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    q075<<-0.3749081498124384,
            0.005475412518938176,
            0.5205190753159222,
            -0.014736681577921108,
            0.008755836621635187,
            -0.05772528980272813,
            0.9981853317186729,
            4.082110261191296e-16,
            -6.141573573129372e-17,
            -4.6355150389749347e-17,
            -3.9899527689833074e-18,
            0.261799388,
            0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.261799388,
            -0.174532925,
            0.0,
            -0.523598776,
            0.0,
            0.0,
            0.17,
            0.11556153466240314,
            0.01590203124693374,
            -1.1725732644862645,
            0.8019999306652887,
            0.35138870088114865,
            0.012513915592962702,
            0.11528210266240224,
            0.001032313722574598,
            0.28542707619682245,
            0.8826828126441494,
            -1.1872920198177317,
            0.027386734120996258,
            0.001,
            0.0,
            0.0,
            0.09999999999999999,
            0.0,
            0.0;

    State s0 = createState(fullBody,q0);
    State s02 = createState(fullBody,q02);
    State s04 = createState(fullBody,q04);
    State s06 = createState(fullBody,q06);
    State s07 = createState(fullBody,q07);
    State s075 = createState(fullBody,q075);
    BOOST_CHECK(reachability::isReachable(fullBody,s0,s02).success());
    BOOST_CHECK(reachability::isReachable(fullBody,s0,s04).success());
    BOOST_CHECK( ! reachability::isReachable(fullBody,s0,s06).success());
    BOOST_CHECK( ! reachability::isReachable(fullBody,s0,s07).success());
    BOOST_CHECK( ! reachability::isReachable(fullBody,s0,s075).success());// should fail in quasi-static

}


BOOST_AUTO_TEST_SUITE_END()

